@@ -399,6 +399,8 @@ typedef enum {
 #define LYD_OPT_CONFIG ...
 #define LYD_OPT_GET ...
 #define LYD_OPT_STRICT ...
+#define LYD_OPT_EDIT ...
+#define LYD_OPT_GETCONFIG ...
 #define LYD_OPT_TRUSTED ...
 #define LYD_OPT_DATA_NO_YANGLIB ...
 #define LYD_OPT_RPC ...
