@@ -5,6 +5,7 @@ import logging
 from typing import IO, Any, Dict, Iterator, Optional, Union
 
 from _libyang import ffi, lib
+from .keyed_list import KeyedList
 from .schema import Module, SContainer, SLeaf, SLeafList, SList, SNode, SRpc, Type
 from .util import LibyangError, c2str, deprecated, str2c
 
@@ -64,15 +65,21 @@ def parser_flags(
     data: bool = False,
     config: bool = False,
     get: bool = False,
+    getconfig: bool = False,
+    edit: bool = False,
+    rpc: bool = False,
+    rpcreply: bool = False,
     strict: bool = False,
     trusted: bool = False,
     no_yanglib: bool = False,
-    rpc: bool = False,
-    rpcreply: bool = False,
     destruct: bool = False,
     no_siblings: bool = False,
     explicit: bool = False,
 ) -> int:
+    if (data, config, get, getconfig, edit, rpc, rpcreply).count(True) > 1:
+        raise ValueError(
+            "Only one of data, config, get, getconfig, edit, rpc, rpcreply can be True"
+        )
     flags = 0
     if data:
         flags |= lib.LYD_OPT_DATA
@@ -80,16 +87,20 @@ def parser_flags(
         flags |= lib.LYD_OPT_CONFIG
     if get:
         flags |= lib.LYD_OPT_GET
+    if getconfig:
+        flags |= lib.LYD_OPT_GETCONFIG
+    if edit:
+        flags |= lib.LYD_OPT_EDIT
+    if rpc:
+        flags |= lib.LYD_OPT_RPC
+    if rpcreply:
+        flags |= lib.LYD_OPT_RPCREPLY
     if strict:
         flags |= lib.LYD_OPT_STRICT
     if trusted:
         flags |= lib.LYD_OPT_TRUSTED
     if no_yanglib:
         flags |= lib.LYD_OPT_DATA_NO_YANGLIB
-    if rpc:
-        flags |= lib.LYD_OPT_RPC
-    if rpcreply:
-        flags |= lib.LYD_OPT_RPCREPLY
     if destruct:
         flags |= lib.LYD_OPT_DESTRUCT
     if no_siblings:
@@ -189,6 +200,8 @@ class DNode:
         data: bool = False,
         config: bool = False,
         get: bool = False,
+        getconfig: bool = False,
+        edit: bool = False,
         rpc: bool = False,
         rpcreply: bool = False,
         no_yanglib: bool = False,
@@ -199,6 +212,8 @@ class DNode:
             data=data,
             config=config,
             get=get,
+            getconfig=getconfig,
+            edit=edit,
             rpc=rpc,
             rpcreply=rpcreply,
             no_yanglib=no_yanglib,
@@ -274,6 +289,29 @@ class DNode:
         if ret != 0:
             raise self.context.error("cannot print node")
 
+    def should_print(
+        self,
+        include_implicit_defaults: bool = False,
+        trim_default_values: bool = False,
+        keep_empty_containers: bool = False,
+    ) -> bool:
+        """
+        Check if a node should be "printed".
+
+        :arg bool include_implicit_defaults:
+            Include implicit default nodes.
+        :arg bool trim_default_values:
+            Exclude nodes with the value equal to their default value.
+        :arg bool keep_empty_containers:
+            Preserve empty non-presence containers.
+        """
+        flags = printer_flags(
+            include_implicit_defaults=include_implicit_defaults,
+            trim_default_values=trim_default_values,
+            keep_empty_containers=keep_empty_containers,
+        )
+        return bool(lib.lyd_node_should_print(self.cdata, flags))
+
     def print_dict(
         self,
         strip_prefixes: bool = True,
@@ -307,27 +345,68 @@ class DNode:
             keep_empty_containers=keep_empty_containers,
         )
 
+        name_cache = {}
+
+        def _node_name(node):
+            name = name_cache.get(node.schema)
+            if name is None:
+                if strip_prefixes:
+                    name = c2str(node.schema.name)
+                else:
+                    mod = lib.lyd_node_module(node)
+                    name = "%s:%s" % (c2str(mod.name), c2str(node.schema.name))
+                name_cache[node.schema] = name
+            return name
+
+        list_keys_cache = {}
+
+        def _init_yang_list(snode):
+            if snode.flags & lib.LYS_USERORDERED:
+                return []  # ordered list, return an empty builtin list
+
+            # unordered lists
+            if snode.nodetype == SNode.LEAFLIST:
+                return KeyedList(key_name=None)
+
+            if snode not in list_keys_cache:
+                list_snode = ffi.cast("struct lys_node_list *", snode)
+                keys = []
+                for i in range(list_snode.keys_size):
+                    key = ffi.cast("struct lys_node *", list_snode.keys[i])
+                    keys.append(c2str(key.name))
+                if len(keys) == 1:
+                    list_keys_cache[snode] = keys[0]
+                else:
+                    list_keys_cache[snode] = tuple(keys)
+
+            return KeyedList(key_name=list_keys_cache[snode])
+
         def _to_dict(node, parent_dic):
-            if not lib.lyd_node_should_print(node.cdata, flags):
+            if not lib.lyd_node_should_print(node, flags):
                 return
-            if strip_prefixes:
-                name = node.name()
-            else:
-                name = "%s:%s" % (node.module().name(), node.name())
-            if isinstance(node, DList):
+            name = _node_name(node)
+            if node.schema.nodetype == SNode.LIST:
                 list_element = {}
-                for child in node:
+                child = node.child
+                while child:
                     _to_dict(child, list_element)
-                parent_dic.setdefault(name, []).append(list_element)
-            elif isinstance(node, (DContainer, DRpc)):
+                    child = child.next
+                if name not in parent_dic:
+                    parent_dic[name] = _init_yang_list(node.schema)
+                parent_dic[name].append(list_element)
+            elif node.schema.nodetype & (SNode.CONTAINER | SNode.RPC | SNode.ACTION):
                 container = {}
-                for child in node:
+                child = node.child
+                while child:
                     _to_dict(child, container)
+                    child = child.next
                 parent_dic[name] = container
-            elif isinstance(node, DLeafList):
-                parent_dic.setdefault(name, []).append(node.value())
-            elif isinstance(node, DLeaf):
-                parent_dic[name] = node.value()
+            elif node.schema.nodetype == SNode.LEAFLIST:
+                if name not in parent_dic:
+                    parent_dic[name] = _init_yang_list(node.schema)
+                parent_dic[name].append(DLeaf.cdata_leaf_value(node))
+            elif node.schema.nodetype == SNode.LEAF:
+                parent_dic[name] = DLeaf.cdata_leaf_value(node)
 
         dic = {}
         dnode = self
@@ -335,19 +414,22 @@ class DNode:
             dnode = dnode.root()
         if with_siblings:
             for sib in dnode.siblings():
-                _to_dict(sib, dic)
+                _to_dict(sib.cdata, dic)
         else:
-            _to_dict(dnode, dic)
+            _to_dict(dnode.cdata, dic)
         return dic
 
     def merge_data_dict(
         self,
         dic: Dict[str, Any],
+        data: bool = False,
+        config: bool = False,
+        get: bool = False,
+        getconfig: bool = False,
+        edit: bool = False,
         rpc: bool = False,
         rpcreply: bool = False,
         strict: bool = False,
-        data: bool = False,
-        config: bool = False,
         no_yanglib: bool = False,
         validate: bool = True,
     ) -> Optional["DNode"]:
@@ -355,25 +437,31 @@ class DNode:
         Merge a python dictionary into this node. The returned value is the first
         created node.
 
-        :arg dict dic:
+        :arg dic:
             The python dictionary to convert.
-        :arg bool rpc:
-            Data represents RPC or action input parameters.
-        :arg bool rpcreply:
-            Data represents RPC or action output parameters.
-        :arg bool strict:
-            Instead of ignoring (with a warning message) data without schema
-            definition, raise an error.
-        :arg bool data:
+        :arg data:
             Complete datastore content with configuration as well as state data. To
             handle possibly missing (but by default required) ietf-yang-library data,
             use no_yanglib=True.
-        :arg bool config:
+        :arg config:
             Complete datastore without state data.
-        :arg bool no_yanglib:
+        :arg get:
+            Data content from a reply message to the NETCONF <get> operation.
+        :arg getconfig:
+            Data content from a reply message to the NETCONF <get-config> operation.
+        :arg edit:
+            Content of the NETCONF <edit-config> config element.
+        :arg rpc:
+            Data represents RPC or action input parameters.
+        :arg rpcreply:
+            Data represents RPC or action output parameters.
+        :arg strict:
+            Instead of ignoring (with a warning message) data without schema
+            definition, raise an error.
+        :arg no_yanglib:
             Ignore (possibly) missing ietf-yang-library data. Applicable only with
             data=True.
-        :arg bool validate:
+        :arg validate:
             If False, do not validate the modified tree before returning. The validation
             is performed on the top of the modified data tree.
         """
@@ -381,11 +469,14 @@ class DNode:
             dic,
             self.module(),
             parent=self,
+            data=data,
+            config=config,
+            get=get,
+            getconfig=getconfig,
+            edit=edit,
             rpc=rpc,
             rpcreply=rpcreply,
             strict=strict,
-            data=data,
-            config=config,
             no_yanglib=no_yanglib,
             validate=validate,
         )
@@ -420,7 +511,11 @@ class DNode:
     @classmethod
     def new(cls, context: "libyang.Context", cdata) -> "DNode":
         cdata = ffi.cast("struct lyd_node *", cdata)
-        nodecls = cls.NODETYPE_CLASS.get(cdata.schema.nodetype, DNode)
+        nodecls = cls.NODETYPE_CLASS.get(cdata.schema.nodetype, None)
+        if nodecls is None:
+            raise NotImplementedError(
+                "node type %s not implemented" % cdata.schema.nodetype
+            )
         return nodecls(context, cdata)
 
 
@@ -446,6 +541,7 @@ class DContainer(DNode):
 
 # -------------------------------------------------------------------------------------
 @DNode.register(SNode.RPC)
+@DNode.register(SNode.ACTION)
 class DRpc(DContainer):
     pass
 
@@ -459,38 +555,28 @@ class DList(DContainer):
 # -------------------------------------------------------------------------------------
 @DNode.register(SNode.LEAF)
 class DLeaf(DNode):
-
-    __slots__ = DNode.__slots__ + ("cdata_leaf",)
-
-    def __init__(self, context: "libyang.Context", cdata):
-        super().__init__(context, cdata)
-        self.cdata_leaf = ffi.cast("struct lyd_node_leaf_list *", cdata)
-
     @property
     def _leaf(self):
         deprecated("_leaf", "cdata_leaf", "2.0.0")
-        return self.cdata_leaf
+        return ffi.cast("struct lyd_node_leaf_list *", self.cdata)
 
     def value(self) -> Any:
-        if self.cdata_leaf.value_type == Type.EMPTY:
-            return None
-        if self.cdata_leaf.value_type in Type.NUM_TYPES:
-            return int(c2str(self.cdata_leaf.value_str))
-        if self.cdata_leaf.value_type in (
-            Type.STRING,
-            Type.BINARY,
-            Type.ENUM,
-            Type.IDENT,
-            Type.BITS,
-        ):
-            return c2str(self.cdata_leaf.value_str)
-        if self.cdata_leaf.value_type == Type.DEC64:
-            return lib.lyd_dec64_to_double(self.cdata)
-        if self.cdata_leaf.value_type == Type.LEAFREF:
-            referenced = DNode.new(self.context, self.cdata_leaf.value.leafref)
-            return referenced.value()
-        if self.cdata_leaf.value_type == Type.BOOL:
-            return bool(self.cdata_leaf.value.bln)
+        return DLeaf.cdata_leaf_value(self.cdata)
+
+    @staticmethod
+    def cdata_leaf_value(cdata) -> Any:
+        cdata = ffi.cast("struct lyd_node_leaf_list *", cdata)
+        val_type = cdata.value_type
+        if val_type in Type.STR_TYPES:
+            return c2str(cdata.value_str)
+        if val_type in Type.NUM_TYPES:
+            return int(c2str(cdata.value_str))
+        if val_type == Type.BOOL:
+            return bool(cdata.value.bln)
+        if val_type == Type.DEC64:
+            return lib.lyd_dec64_to_double(ffi.cast("struct lyd_node *", cdata))
+        if val_type == Type.LEAFREF:
+            return DLeaf.cdata_leaf_value(cdata.value.leafref)
         return None
 
 
@@ -505,11 +591,14 @@ def dict_to_dnode(
     dic: Dict[str, Any],
     module: Module,
     parent: Optional[DNode] = None,
+    data: bool = False,
+    config: bool = False,
+    get: bool = False,
+    getconfig: bool = False,
+    edit: bool = False,
     rpc: bool = False,
     rpcreply: bool = False,
     strict: bool = False,
-    data: bool = False,
-    config: bool = False,
     no_yanglib: bool = False,
     validate: bool = True,
 ) -> Optional[DNode]:
@@ -524,6 +613,18 @@ def dict_to_dnode(
     :arg parent:
         Optional parent to update. If not specified a new top-level DNode will be
         created.
+    :arg data:
+        Complete datastore content with configuration as well as state data. To handle
+        possibly missing (but by default required) ietf-yang-library data, use
+        no_yanglib=True.
+    :arg config:
+        Complete datastore without state data.
+    :arg get:
+        Data content from a reply message to the NETCONF <get> operation.
+    :arg getconfig:
+        Data content from a reply message to the NETCONF <get-config> operation.
+    :arg edit:
+        Content of the NETCONF <edit-config> config element.
     :arg rpc:
         Data represents RPC or action input parameters.
     :arg rpcreply:
@@ -531,12 +632,6 @@ def dict_to_dnode(
     :arg strict:
         Instead of ignoring (with a warning message) data without schema definition,
         raise an error.
-    :arg data:
-        Complete datastore content with configuration as well as state data. To handle
-        possibly missing (but by default required) ietf-yang-library data, use
-        no_yanglib=True.
-    :arg config:
-        Complete datastore without state data.
     :arg no_yanglib:
         Ignore (possibly) missing ietf-yang-library data. Applicable only with
         data=True.
@@ -718,10 +813,13 @@ def dict_to_dnode(
             result = DNode.new(module.context, created[0])
             if validate:
                 result.root().validate(
-                    rpc=rpc,
-                    rpcreply=rpcreply,
                     data=data,
                     config=config,
+                    get=get,
+                    getconfig=getconfig,
+                    edit=edit,
+                    rpc=rpc,
+                    rpcreply=rpcreply,
                     no_yanglib=no_yanglib,
                 )
     except:
