@@ -129,11 +129,14 @@ class Module:
     def parse_data_dict(
         self,
         dic: Dict[str, Any],
+        data: bool = False,
+        config: bool = False,
+        get: bool = False,
+        getconfig: bool = False,
+        edit: bool = False,
         rpc: bool = False,
         rpcreply: bool = False,
         strict: bool = False,
-        data: bool = False,
-        config: bool = False,
         no_yanglib: bool = False,
         validate: bool = True,
     ) -> "libyang.data.DNode":
@@ -144,6 +147,18 @@ class Module:
 
         :arg dic:
             The python dictionary to convert.
+        :arg data:
+            Complete datastore content with configuration as well as state data. To
+            handle possibly missing (but by default required) ietf-yang-library data,
+            use no_yanglib=True.
+        :arg config:
+            Complete datastore without state data.
+        :arg get:
+            Data content from a reply message to the NETCONF <get> operation.
+        :arg getconfig:
+            Data content from a reply message to the NETCONF <get-config> operation.
+        :arg edit:
+            Content of the NETCONF <edit-config> config element.
         :arg rpc:
             Data represents RPC or action input parameters.
         :arg rpcreply:
@@ -151,12 +166,6 @@ class Module:
         :arg strict:
             Instead of ignoring (with a warning message) data without schema definition,
             raise an error.
-        :arg data:
-            Complete datastore content with configuration as well as state data. To
-            handle possibly missing (but by default required) ietf-yang-library data,
-            use no_yanglib=True.
-        :arg config:
-            Complete datastore without state data.
         :arg no_yanglib:
             Ignore (possibly) missing ietf-yang-library data. Applicable only with
             data=True.
@@ -169,11 +178,14 @@ class Module:
             dic,
             self,
             parent=None,
+            data=data,
+            config=config,
+            get=get,
+            getconfig=getconfig,
+            edit=edit,
             rpc=rpc,
             rpcreply=rpcreply,
             strict=strict,
-            data=data,
-            config=config,
             no_yanglib=no_yanglib,
             validate=validate,
         )
@@ -425,6 +437,8 @@ class Type:
             if rng is not None:
                 yield rng
 
+    STR_TYPES = frozenset((STRING, BINARY, ENUM, IDENT, BITS))
+
     def length(self) -> Optional[str]:
         if self.cdata.base == self.STRING and self.cdata.info.str.length:
             return c2str(self.cdata.info.str.length.expr)
@@ -712,6 +726,7 @@ class SNode:
     LEAFLIST = lib.LYS_LEAFLIST
     LIST = lib.LYS_LIST
     RPC = lib.LYS_RPC
+    ACTION = lib.LYS_ACTION
     INPUT = lib.LYS_INPUT
     OUTPUT = lib.LYS_OUTPUT
     KEYWORDS = {
@@ -720,6 +735,7 @@ class SNode:
         LEAFLIST: "leaf-list",
         LIST: "list",
         RPC: "rpc",
+        ACTION: "action",
         INPUT: "input",
         OUTPUT: "output",
     }
@@ -831,17 +847,20 @@ class SNode:
 
     NODETYPE_CLASS = {}
 
-    @classmethod
-    def register(cls, nodetype):
+    @staticmethod
+    def register(nodetype):
         def _decorator(nodeclass):
-            cls.NODETYPE_CLASS[nodetype] = nodeclass
+            SNode.NODETYPE_CLASS[nodetype] = nodeclass
             return nodeclass
 
         return _decorator
 
-    @classmethod
-    def new(cls, context: "libyang.Context", cdata) -> "SNode":
-        nodecls = cls.NODETYPE_CLASS.get(cdata.nodetype, SNode)
+    @staticmethod
+    def new(context: "libyang.Context", cdata) -> "SNode":
+        cdata = ffi.cast("struct lys_node *", cdata)
+        nodecls = SNode.NODETYPE_CLASS.get(cdata.nodetype, None)
+        if nodecls is None:
+            raise NotImplementedError("node type %s not implemented" % cdata.nodetype)
         return nodecls(context, cdata)
 
 
@@ -999,6 +1018,7 @@ class SRpcInOut(SNode):
 
 # -------------------------------------------------------------------------------------
 @SNode.register(SNode.RPC)
+@SNode.register(SNode.ACTION)
 class SRpc(SNode):
     def input(self) -> Optional[SRpcInOut]:
         try:
@@ -1043,6 +1063,7 @@ def iter_children(
 ) -> Iterator[SNode]:
     if types is None:
         types = (
+            lib.LYS_ACTION,
             lib.LYS_CONTAINER,
             lib.LYS_LIST,
             lib.LYS_RPC,
